/***************************************************************************
 *   Copyright (C) 2008 by root   *
 *   root@skystar   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "../Includes/qstdinc.h"
#include "../Includes/qdvbts.h"
#include "../Includes/qcrc.h"
#include "qdvbts_priv.h"

bool QDVBValidCRC(uint8_t* pData, uint32_t nSize)
{
	// Create CRC32

	if (nSize < 4) return false;
    uint32_t nCRC32 = QCalcCRC(pData, nSize - 4);

	uint32_t nCRCOrg = 0;
	nCRCOrg = pData[nSize - 4];  nCRCOrg = nCRCOrg << 8;
	nCRCOrg = nCRCOrg | ((uint32_t)(pData[nSize - 3])); nCRCOrg = nCRCOrg << 8;
	nCRCOrg = nCRCOrg | ((uint32_t)(pData[nSize - 2])); nCRCOrg = nCRCOrg << 8;
	nCRCOrg = nCRCOrg | ((uint32_t)(pData[nSize - 1]));

	return nCRCOrg == nCRC32;
}

bool QDVBParseTS(uint8_t* pData, QDVBTSPacket* pTS)
{
	QDVB_Start(pData, 188);

	memset(pTS, 0, sizeof(QDVBTSPacket));

	// Common
	QDVB_ReadBits(pTS->sync_byte, 8);
	QDVB_ReadBits(pTS->transport_error_indicator, 1);
	QDVB_ReadBits(pTS->payload_unit_start_indicator, 1);
	QDVB_ReadBits(pTS->transport_priority, 1);
	QDVB_ReadBits(pTS->PID, 13);
	QDVB_ReadBits(pTS->transport_scrambling_control, 2);
	QDVB_ReadBits(pTS->adaptation_field_control, 2);
	QDVB_ReadBits(pTS->continuity_counter, 4);

	// Read adaptation here
	if(pTS->adaptation_field_control & 0x2)
	{
		QDVB_ReadBits(pTS->adaptation_field_length, 8);
		long nCurIndex = dsIndex;
		if (pTS->adaptation_field_length > 0)
		{
			if (pTS->adaptation_field_control == 0x2)
			{
				if (pTS->adaptation_field_length != 183 ) QDVB_GotError();
			}

			if (pTS->adaptation_field_control == 0x3)
			{
				if (pTS->adaptation_field_length > 182 ) QDVB_GotError();
			}

			QDVB_ReadBits(pTS->discontinuity_indicator, 1);
			QDVB_ReadBits(pTS->random_access_indicator, 1);
			QDVB_ReadBits(pTS->elementary_stream_priority_indicator, 1);
			QDVB_ReadBits(pTS->PCR_flag, 1);
			QDVB_ReadBits(pTS->OPCR_flag, 1);
			QDVB_ReadBits(pTS->splicing_point_flag, 1);
			QDVB_ReadBits(pTS->transport_private_data_flag, 1);
			QDVB_ReadBits(pTS->adaptation_field_extension_flag, 1);
			// PCR
			if (pTS->PCR_flag == 1)
			{
				QDVB_ReadBits(pTS->program_clock_reference_base, 33);
				QDVB_SkipBits(6);
				QDVB_ReadBits(pTS->program_clock_reference_extension, 9);
			}

			// OPCR
			if (pTS->OPCR_flag == 1)
			{
				QDVB_ReadBits(pTS->original_program_clock_reference_base, 33);
				QDVB_SkipBits(6);
				QDVB_ReadBits(pTS->original_program_clock_reference_extension, 9);
			}

			// Splice
			if (pTS->splicing_point_flag == 1)
			{
				QDVB_ReadBits(pTS->splice_countdown, 8);
			}

			// Private
			if (pTS->transport_private_data_flag == 1)
			{
				QDVB_ReadBits(pTS->transport_private_data_length, 8);
				pTS->private_data_byte = QDVB_CurrentBuffer();
				QDVB_SkipBits(8 * pTS->transport_private_data_length);
			}

			// Extension
			if (pTS->adaptation_field_extension_flag == 1)
			{
				QDVB_ReadBits(pTS->adaptation_field_extension_length, 8);
				QDVB_ReadBits(pTS->ltw_flag, 1);
				QDVB_ReadBits(pTS->piecewise_rate_flag, 1);
				QDVB_ReadBits(pTS->seamless_splice_flag, 1);
				QDVB_SkipBits(5);

				// LTW
				if (pTS->ltw_flag == 1) 
				{
					QDVB_ReadBits(pTS->ltw_valid_flag, 1);
					QDVB_ReadBits(pTS->ltw_offset, 15);
				}

				// Piecewise
				if (pTS->piecewise_rate_flag == 1) 
				{
					QDVB_SkipBits(2);
					QDVB_ReadBits(pTS->piecewise_rate, 22);
				}
				
				// Seamless splices
				if (pTS->seamless_splice_flag == 1) 
				{
					QDVB_ReadBits(pTS->splice_type, 4);
					QDVB_ReadBits(pTS->DTS_next_AU, 3);
					QDVB_SkipBits(1);
					QDVB_ReadBits(pTS->DTS_next_AU, 15);
					QDVB_SkipBits(1);
					QDVB_ReadBits(pTS->DTS_next_AU, 15);
					QDVB_SkipBits(1);
				}

			}
		}

		QDVB_RoundNextByte();
		dsIndex = nCurIndex + (pTS->adaptation_field_length);
	}

	// Fix buffer position
	pTS->data = QDVB_CurrentBuffer();
	pTS->data_size = (uint8_t)QDVB_CurrentSize();

	QDVB_BoundsError;
	return QDVB_BCError();
}

bool QDVBMakeTS(uint8_t* pData, QDVBTSPacket* pTS)
{
	long nTmp = 0;
	QDVB_StartWrite(pData, nTmp);

	// Common
	QDVB_WriteBits(pTS->sync_byte, 8);
	QDVB_WriteBits(pTS->transport_error_indicator, 1);
	QDVB_WriteBits(pTS->payload_unit_start_indicator, 1);
	QDVB_WriteBits(pTS->transport_priority, 1);
	QDVB_WriteBits(pTS->PID, 13);
	QDVB_WriteBits(pTS->transport_scrambling_control, 2);
	QDVB_WriteBits(pTS->adaptation_field_control, 2);
	QDVB_WriteBits(pTS->continuity_counter, 4);

	// Read adaptation here
	if (pTS->adaptation_field_control & 0x2)
	{
		QDVB_WriteBits(pTS->adaptation_field_length, 8);
		long nCurIndex = dsIndex;
		if (pTS->adaptation_field_length > 0)
		{
			QDVB_WriteBits(pTS->discontinuity_indicator, 1);
			QDVB_WriteBits(pTS->random_access_indicator, 1);
			QDVB_WriteBits(pTS->elementary_stream_priority_indicator, 1);
			QDVB_WriteBits(pTS->PCR_flag, 1);
			QDVB_WriteBits(pTS->OPCR_flag, 1);
			QDVB_WriteBits(pTS->splicing_point_flag, 1);
			QDVB_WriteBits(pTS->transport_private_data_flag, 1);
			QDVB_WriteBits(pTS->adaptation_field_extension_flag, 1);
			// PCR
			if (pTS->PCR_flag == 1)
			{
				QDVB_WriteBits(pTS->program_clock_reference_base, 33);
				QDVB_WriteBits(0, 6);
				QDVB_WriteBits(pTS->program_clock_reference_extension, 9);
			}

			// OPCR
			if (pTS->OPCR_flag == 1)
			{
				QDVB_WriteBits(pTS->original_program_clock_reference_base, 33);
				QDVB_WriteBits(0, 6);
				QDVB_WriteBits(pTS->original_program_clock_reference_extension,
						9);
			}

			// Splice
			if (pTS->splicing_point_flag == 1)
			{
				QDVB_WriteBits(pTS->splice_countdown, 8);
			}

			// Private
			if (pTS->transport_private_data_flag == 1)
			{
				QDVB_WriteBits(pTS->transport_private_data_length, 8);
				for (int i = 0; i < pTS->transport_private_data_length; i++)
				{
					QDVB_WriteBits(pTS->private_data_byte[i], 8);
				}
			}

			// Extension
			if (pTS->adaptation_field_extension_flag == 1)
			{
				QDVB_WriteBits(pTS->adaptation_field_extension_length, 8);
				QDVB_WriteBits(pTS->ltw_flag, 1);
				QDVB_WriteBits(pTS->piecewise_rate_flag, 1);
				QDVB_WriteBits(pTS->seamless_splice_flag, 1);
				QDVB_WriteBits(0, 5);

				// LTW
				if (pTS->ltw_flag == 1)
				{
					QDVB_WriteBits(pTS->ltw_valid_flag, 1);
					QDVB_WriteBits(pTS->ltw_offset, 15);
				}

				// Piecewise
				if (pTS->piecewise_rate_flag == 1)
				{
					QDVB_WriteBits(0, 2);
					QDVB_WriteBits(pTS->piecewise_rate, 22);
				}

				// Seamless splices
				if (pTS->seamless_splice_flag == 1)
				{
					QDVB_WriteBits(pTS->splice_type, 4);
					QDVB_WriteBits(pTS->DTS_next_AU >> 30, 3);
					QDVB_WriteBits(0, 1);
					QDVB_WriteBits(pTS->DTS_next_AU >> 15, 15);
					QDVB_WriteBits(0, 1);
					QDVB_WriteBits(pTS->DTS_next_AU, 15);
					QDVB_WriteBits(0, 1);
				}

			}
		}

		QDVB_RoundNextByte();
		dsIndex = nCurIndex + (pTS->adaptation_field_length);
	}

	// Fix buffer position
	for (int i = 0; i < pTS->data_size; i++)
	{
		QDVB_WriteBits(pTS->data[i], 8);
	}

	return QDVB_BCError();
}

bool QDVBPatchPCR(uint8_t* pData, QDVBTSPacket* pTS)
{
	long nTmp = 0;
	QDVB_StartWrite(pData, nTmp);

	// Common
	QDVB_WriteBits(pTS->sync_byte, 8);
	QDVB_WriteBits(pTS->transport_error_indicator, 1);
	QDVB_WriteBits(pTS->payload_unit_start_indicator, 1);
	QDVB_WriteBits(pTS->transport_priority, 1);
	QDVB_WriteBits(pTS->PID, 13);
	QDVB_WriteBits(pTS->transport_scrambling_control, 2);
	QDVB_WriteBits(pTS->adaptation_field_control, 2);
	QDVB_WriteBits(pTS->continuity_counter, 4);

	if (pTS->adaptation_field_control & 0x2)
	{
		QDVB_WriteBits(pTS->adaptation_field_length, 8);
		if (pTS->adaptation_field_length > 0)
		{
			QDVB_WriteBits(pTS->discontinuity_indicator, 1);
			QDVB_WriteBits(pTS->random_access_indicator, 1);
			QDVB_WriteBits(pTS->elementary_stream_priority_indicator, 1);
			QDVB_WriteBits(pTS->PCR_flag, 1);
			QDVB_WriteBits(pTS->OPCR_flag, 1);
			QDVB_WriteBits(pTS->splicing_point_flag, 1);
			QDVB_WriteBits(pTS->transport_private_data_flag, 1);
			QDVB_WriteBits(pTS->adaptation_field_extension_flag, 1);

			// PCR
			if (pTS->PCR_flag == 1)
			{
				QDVB_WriteBits(pTS->program_clock_reference_base, 33);
			}
		}
	}

	return QDVB_BCError();
}

void QDVBPatchCC(uint8_t* pData, int nCC)
{
	pData[0x03] = (pData[0x03] & 0xf0) | (nCC & 0x0f);
}


bool QDVBParsePES(uint8_t* pData, int nSize, QDVBPESHdr* pPES, bool bHdrOnly)
{
	QDVB_Start(pData, nSize);
	uint8_t temp = 0;
	int i = 0;
	uint8_t* pExtPtr = NULL;
	uint64_t nComb[3];

	memset(pPES, 0, sizeof(QDVBPESHdr));

	// Common
	QDVB_ReadBits(pPES->packet_start_code_prefix, 24);
	QDVB_ReadBits(pPES->stream_id, 8);
	QDVB_ReadBits(pPES->PES_packet_length, 16);

	QDVB_ReadBits(temp, 2);									// 2 bslbf '10'
	QDVB_ReadBits(pPES->PES_scrambling_control, 2);			// 2 bslbf
	QDVB_ReadBits(pPES->PES_priority, 1);					// 1 bslbf
	QDVB_ReadBits(pPES->data_alignment_indicator, 1);		// 1 bslbf
	QDVB_ReadBits(pPES->copyright, 1);						// 1 bslbf
	QDVB_ReadBits(pPES->original_or_copy, 1);				// 1 bslbf
	QDVB_ReadBits(pPES->PTS_DTS_flags, 2);					// 2 bslbf
	QDVB_ReadBits(pPES->ESCR_flag, 1);						// 1 bslbf
	QDVB_ReadBits(pPES->ES_rate_flag, 1);					// 1 bslbf
	QDVB_ReadBits(pPES->DSM_trick_mode_flag, 1);				// 1 bslbf
	QDVB_ReadBits(pPES->additional_copy_info_flag, 1);		// 1 bslbf
	QDVB_ReadBits(pPES->PES_CRC_flag, 1);					// 1 bslbf
	QDVB_ReadBits(pPES->PES_extension_flag, 1);				// 1 bslbf
	QDVB_ReadBits(pPES->PES_header_data_length, 8);			// 8 uimsbf
	pExtPtr = QDVB_CurrentBuffer();

	// DTS modes
	if (pPES->PTS_DTS_flags == 2)
	{
		nComb[2] = nComb[1] = nComb[0] = 0;
		QDVB_ReadBits(temp, 4);
		QDVB_ReadBits(nComb[2], 3);
		QDVB_SkipBits(1);
		QDVB_ReadBits(nComb[1], 15);
		QDVB_SkipBits(1);
		QDVB_ReadBits(nComb[0], 15);
		QDVB_SkipBits(1);
		pPES->PTS = nComb[0] | (nComb[1] << 15) | (nComb[2] << 30);
	}
	else if (pPES->PTS_DTS_flags == 3)
	{
		nComb[2] = nComb[1] = nComb[0] = 0;
		QDVB_ReadBits(temp, 4);
		QDVB_ReadBits(nComb[2], 3);
		QDVB_SkipBits(1);
		QDVB_ReadBits(nComb[1], 15);
		QDVB_SkipBits(1);
		QDVB_ReadBits(nComb[0], 15);
		QDVB_SkipBits(1);
		pPES->PTS = nComb[0] | (nComb[1] << 15) | (nComb[2] << 30);

		nComb[2] = nComb[1] = nComb[0] = 0;
		QDVB_ReadBits(temp, 4);
		QDVB_ReadBits(nComb[2], 3);
		QDVB_SkipBits(1);
		QDVB_ReadBits(nComb[1], 15);
		QDVB_SkipBits(1);
		QDVB_ReadBits(nComb[0], 15);
		QDVB_SkipBits(1);
		pPES->DTS = nComb[0] | (nComb[1] << 15) | (nComb[2] << 30);
	}

	if (bHdrOnly) return (pPES->packet_start_code_prefix == 1);

	// ESCR
	if (pPES->ESCR_flag == 1)
	{
		nComb[2] = nComb[1] = nComb[0] = 0;
		QDVB_ReadBits(temp, 2);
		QDVB_ReadBits(nComb[2], 3);
		QDVB_SkipBits(1);
		QDVB_ReadBits(nComb[1], 15);
		QDVB_SkipBits(1);
		QDVB_ReadBits(nComb[0], 15);
		QDVB_SkipBits(1);
		QDVB_ReadBits(pPES->ESCR_ext, 9);
		QDVB_SkipBits(1);
		pPES->ESCR_base = nComb[0] | (nComb[1] << 15) | (nComb[2] << 30);
	}

	// ESCR rate
	if (pPES->ES_rate_flag == 1)
	{
		QDVB_SkipBits(1);
		QDVB_ReadBits(pPES->ES_rate, 22);
		QDVB_SkipBits(1);
	}

	// DSM_trick_mode_flag
	if (pPES->DSM_trick_mode_flag == 1)
	{
		QDVB_ReadBits(pPES->trick_mode_control, 3);
		if (pPES->trick_mode_control == 0)
		{
			QDVB_ReadBits(pPES->field_id, 2);
			QDVB_ReadBits(pPES->intra_slice_refresh, 1);
			QDVB_ReadBits(pPES->frequency_truncation, 2);
		}
		else if (pPES->trick_mode_control == 1)
		{
			QDVB_ReadBits(pPES->rep_cntrl, 5);
		}
		else if (pPES->trick_mode_control == 2)
		{
			QDVB_ReadBits(pPES->field_id, 2);
			QDVB_SkipBits(3);
		}
		else if (pPES->trick_mode_control == 3)
		{
			QDVB_ReadBits(pPES->field_id, 2);
			QDVB_ReadBits(pPES->intra_slice_refresh, 1);
			QDVB_ReadBits(pPES->frequency_truncation, 2);
		}
		else if (pPES->trick_mode_control == 4)
		{
			QDVB_ReadBits(pPES->rep_cntrl, 5);
		}
		else
		{
			QDVB_ReadBits(temp, 5);
		}
	}

	if (pPES->additional_copy_info_flag == 1)
	{
		QDVB_SkipBits(1);
		QDVB_ReadBits(pPES->additional_copy_info, 7);
	}

	if (pPES->PES_CRC_flag == 1)
	{
		QDVB_ReadBits(pPES->previous_PES_packet_CRC, 16);
	}

	if (pPES->PES_extension_flag == 1)
	{
		QDVB_ReadBits(pPES->PES_private_data_flag, 1);
		QDVB_ReadBits(pPES->pack_header_field_flag, 1);
		QDVB_ReadBits(pPES->program_packet_sequence_counter_flag, 1);
		QDVB_ReadBits(pPES->P_STD_buffer_flag, 1);
		QDVB_SkipBits(3);
		QDVB_ReadBits(pPES->PES_extension_flag_2, 1);

		if (pPES->PES_private_data_flag == 1)
		{
			for (int i=0; i<16; i++)
			{
				QDVB_ReadBits(pPES->private_data[i], 8);
			}
		}

		if (pPES->pack_header_field_flag == 1)
		{
			QDVB_ReadBits(pPES->pack_field_length, 8);
			for (int i=0; i<pPES->pack_field_length; i++)
			{
				QDVB_ReadBits(pPES->pack_field[i], 8);
			}
		}

		if (pPES->program_packet_sequence_counter_flag == 1)
		{
			QDVB_SkipBits(1);
			QDVB_ReadBits(pPES->program_packet_sequence_counter, 7);
			QDVB_SkipBits(1);
			QDVB_ReadBits(pPES->MPEG1_MPEG2_identifier, 1);
			QDVB_ReadBits(pPES->original_stuff_length, 6);
		}

		if (pPES->P_STD_buffer_flag == 1)
		{
			QDVB_SkipBits(2);
			QDVB_ReadBits(pPES->P_STD_buffer_scale, 1);
			QDVB_ReadBits(pPES->P_STD_buffer_size, 13);
		}

		if (pPES->PES_extension_flag_2 == 1)
		{
			QDVB_SkipBits(1);
			QDVB_ReadBits(pPES->PES_extension_field_length, 7);
			for (i = 0; i < pPES->PES_extension_field_length; i++)
			{
				QDVB_ReadBits(pPES->PES_extension_field[i], 8);
			}
		}
	}

	if (!bHdrOnly)
	{
		while((QDVB_CurrentBuffer() - pExtPtr) < pPES->PES_header_data_length)
		{
			// Skip stuffing bytes
			QDVB_SkipBits(8);
		}

		// Data is here
		pPES->data = QDVB_CurrentBuffer();
		pPES->data_size = QDVB_CurrentSize();
	}

QDVB_BoundsError;
	return QDVB_BCError() && (pPES->packet_start_code_prefix == 1);
}

bool QDVBCreatePES(uint8_t* pBuffer, long& nBufferSize, uint8_t* pData, long nDataSize, QDVBPESHdr* pPES, bool bUnlim)
{
	QDVB_StartWrite(pBuffer, nBufferSize);
	uint8_t* pHdrPos = NULL;
	uint8_t* pLenPos = NULL;
	uint16_t nComb[3];
	int nHdrLen = 0;
	int nPESLen = 0;

	// Common
	QDVB_WriteBits(pPES->packet_start_code_prefix, 24);
	QDVB_WriteBits(pPES->stream_id, 8);
	pLenPos = QDVB_CurrentBuffer();
	QDVB_WriteBits(0, 16);

	QDVB_WriteBits(2, 2);									// 2 bslbf '10'
	QDVB_WriteBits(pPES->PES_scrambling_control, 2);			// 2 bslbf
	QDVB_WriteBits(pPES->PES_priority, 1);					// 1 bslbf
	QDVB_WriteBits(pPES->data_alignment_indicator, 1);		// 1 bslbf
	QDVB_WriteBits(pPES->copyright, 1);						// 1 bslbf
	QDVB_WriteBits(pPES->original_or_copy, 1);				// 1 bslbf
	QDVB_WriteBits(pPES->PTS_DTS_flags, 2);					// 2 bslbf
	QDVB_WriteBits(pPES->ESCR_flag, 1);						// 1 bslbf
	QDVB_WriteBits(pPES->ES_rate_flag, 1);					// 1 bslbf
	QDVB_WriteBits(pPES->DSM_trick_mode_flag, 1);			// 1 bslbf
	QDVB_WriteBits(pPES->additional_copy_info_flag, 1);		// 1 bslbf
	QDVB_WriteBits(pPES->PES_CRC_flag, 1);					// 1 bslbf
	QDVB_WriteBits(pPES->PES_extension_flag, 1);				// 1 bslbf
	pHdrPos = QDVB_CurrentBuffer();
	QDVB_WriteBits(0, 8);									// 8 uimsbf

	// DTS modes
	if (pPES->PTS_DTS_flags == 2)
	{
		QDVB_WriteBits(2, 4);				// 4 bslbf

		nComb[0] = pPES->PTS & 0x7fff;
		nComb[1] = (pPES->PTS >> 15) & 0x7fff;
		nComb[2] = (pPES->PTS >> 30) & 0x7;

		QDVB_WriteBits(nComb[2], 3);
		QDVB_WriteBits(1, 1);
		QDVB_WriteBits(nComb[1], 15);
		QDVB_WriteBits(1, 1);
		QDVB_WriteBits(nComb[0], 15);
		QDVB_WriteBits(1, 1);
	}
	else if (pPES->PTS_DTS_flags == 3)
	{
		QDVB_WriteBits(3, 4);				// 4 bslbf

		nComb[0] = pPES->PTS & 0x7fff;
		nComb[1] = (pPES->PTS >> 15) & 0x7fff;
		nComb[2] = (pPES->PTS >> 30) & 0x7;

		QDVB_WriteBits(nComb[2], 3);
		QDVB_WriteBits(1, 1);
		QDVB_WriteBits(nComb[1], 15);
		QDVB_WriteBits(1, 1);
		QDVB_WriteBits(nComb[0], 15);
		QDVB_WriteBits(1, 1);

		QDVB_WriteBits(1, 4);				// 4 bslbf

		nComb[0] = pPES->DTS & 0x7fff;
		nComb[1] = (pPES->DTS >> 15) & 0x7fff;
		nComb[2] = (pPES->DTS >> 30) & 0x7;

		QDVB_WriteBits(nComb[2], 3);
		QDVB_WriteBits(1, 1);
		QDVB_WriteBits(nComb[1], 15);
		QDVB_WriteBits(1, 1);
		QDVB_WriteBits(nComb[0], 15);
		QDVB_WriteBits(1, 1);
	}

	// ESCR
	if (pPES->ESCR_flag == 1)
	{
		QDVB_WriteBits(0, 2);

		nComb[0] = pPES->ESCR_base & 0x7fff;
		nComb[1] = (pPES->ESCR_base >> 15) & 0x7fff;
		nComb[2] = (pPES->ESCR_base >> 30) & 0x7;

		QDVB_WriteBits(nComb[2], 3);
		QDVB_WriteBits(1, 1);
		QDVB_WriteBits(nComb[1], 15);
		QDVB_WriteBits(1, 1);
		QDVB_WriteBits(nComb[0], 15);
		QDVB_WriteBits(1, 1);
		QDVB_WriteBits(pPES->ESCR_ext, 9);
		QDVB_WriteBits(1, 1);
	}

	// ESCR rate
	if (pPES->ES_rate_flag == 1)
	{
		QDVB_WriteBits(1, 1);
		QDVB_WriteBits(pPES->ES_rate, 22);
		QDVB_WriteBits(1, 1);
	}

	// DSM_trick_mode_flag
	if (pPES->DSM_trick_mode_flag == 1)
	{
		QDVB_WriteBits(pPES->trick_mode_control, 3);
		if (pPES->trick_mode_control == 0)
		{
			QDVB_WriteBits(pPES->field_id, 2);
			QDVB_WriteBits(pPES->intra_slice_refresh, 1);
			QDVB_WriteBits(pPES->frequency_truncation, 2);
		}
		else if (pPES->trick_mode_control == 1)
		{
			QDVB_WriteBits(pPES->rep_cntrl, 5);
		}
		else if (pPES->trick_mode_control == 2)
		{
			QDVB_WriteBits(pPES->field_id, 2);
			QDVB_WriteBits(0, 3);
		}
		else if (pPES->trick_mode_control == 3)
		{
			QDVB_WriteBits(pPES->field_id, 2);
			QDVB_WriteBits(pPES->intra_slice_refresh, 1);
			QDVB_WriteBits(pPES->frequency_truncation, 2);
		}
		else if (pPES->trick_mode_control == 4)
		{
			QDVB_WriteBits(pPES->rep_cntrl, 5);
		}
		else
		{
			QDVB_WriteBits(0, 5);
		}
	}

	if (pPES->additional_copy_info_flag == 1)
	{
		QDVB_WriteBits(1, 1);
		QDVB_WriteBits(pPES->additional_copy_info, 7);
	}

	if (pPES->PES_CRC_flag == 1)
	{
		QDVB_WriteBits(pPES->previous_PES_packet_CRC, 16);
	}

	if (pPES->PES_extension_flag == 1)
	{
		QDVB_WriteBits(pPES->PES_private_data_flag, 1);
		QDVB_WriteBits(pPES->pack_header_field_flag, 1);
		QDVB_WriteBits(pPES->program_packet_sequence_counter_flag, 1);
		QDVB_WriteBits(pPES->P_STD_buffer_flag, 1);
		QDVB_WriteBits(0, 3);
		QDVB_WriteBits(pPES->PES_extension_flag_2, 1);

		if (pPES->PES_private_data_flag == 1)
		{
			for (int i=0; i<16; i++)
			{
				QDVB_WriteBits(pPES->private_data[i], 8);
			}
		}

		if (pPES->pack_header_field_flag == 1)
		{
			QDVB_WriteBits(pPES->pack_field_length, 8);
			for (int i=0; i<pPES->pack_field_length; i++)
			{
				QDVB_WriteBits(pPES->pack_field[i], 8);
			}
		}

		if (pPES->program_packet_sequence_counter_flag == 1)
		{
			QDVB_WriteBits(1, 1);
			QDVB_WriteBits(pPES->program_packet_sequence_counter, 7);
			QDVB_WriteBits(1, 1);
			QDVB_WriteBits(pPES->MPEG1_MPEG2_identifier, 1);
			QDVB_WriteBits(pPES->original_stuff_length, 6);
		}

		if (pPES->P_STD_buffer_flag == 1)
		{
			QDVB_WriteBits(1, 2);
			QDVB_WriteBits(pPES->P_STD_buffer_scale, 1);
			QDVB_WriteBits(pPES->P_STD_buffer_size, 13);
		}

		if (pPES->PES_extension_flag_2 == 1)
		{
			QDVB_WriteBits(1, 1);
			QDVB_WriteBits(pPES->PES_extension_field_length, 7);
			for (int i = 0; i < pPES->PES_extension_field_length; i++)
			{
				QDVB_WriteBits(pPES->PES_extension_field[i], 8);
			}
		}
	}

	nHdrLen = (int)(QDVB_CurrentBuffer() - pHdrPos - 1);
	*pHdrPos = nHdrLen;
	for (int i=0; i<nDataSize; i++)
	{
		QDVB_WriteBits(pData[i], 8);
	}

	uint16_t nOutLen = 0;
	nPESLen = (int)(QDVB_CurrentBuffer() - pBuffer - 6); // Not count starting 6 bytes
	if ((nPESLen < 0xFFFF) && (!bUnlim))
	{
		nOutLen = nPESLen;
	}

	pLenPos[0] = (nOutLen >> 8) & 0xff;
	pLenPos[1] = (nOutLen) & 0xff;

	nBufferSize = nPESLen + 6;

	return QDVB_BCError();
}

QDVBDescCA::QDVBDescCA()
{
	/* Preinit vars */
	descriptor_tag = 0;
    CA_system_ID = (uint16_t)-1;
    CA_PID = (uint16_t)-1;
}

QDVBDescCA::QDVBDescCA(QDVBDesc* pSrc)
{
	/* Preinit vars */
	descriptor_tag = 0;
    CA_system_ID = (uint16_t)-1;
    CA_PID = (uint16_t)-1;

    /* Parse source */
    if (pSrc->descriptor_tag != 9) return;
    descriptor_tag = pSrc->descriptor_tag;

	QDVB_Start(pSrc->data.data(), pSrc->data.size());
	QDVB_ReadBits(CA_system_ID, 16);
	QDVB_SkipBits(3);
	QDVB_ReadBits(CA_PID, 13);
    data.resize(QDVB_CurrentSize());
    memcpy(data.data(), QDVB_CurrentBuffer(), QDVB_CurrentSize());

QDVB_BoundsError;
	return;
}

QDVBDescCA::~QDVBDescCA()
{

}

QDVBDescCA* QDVBDescCA::Duplicate()
{
	QDVBDescCA* pDst = new QDVBDescCA();

	pDst->descriptor_tag = descriptor_tag;
	pDst->CA_system_ID = CA_system_ID;
    pDst->CA_PID = CA_PID;

	pDst->data.resize(data.size());
	memcpy(pDst->data.data(), data.data(), data.size());
	return pDst;
}

QDVBDesc::QDVBDesc()
{
	descriptor_tag = 0;
}

QDVBDesc::~QDVBDesc()
{
}

QDVBDesc* QDVBDesc::Duplicate()
{
	QDVBDesc* pRetVal = new QDVBDesc();
	pRetVal->descriptor_tag = descriptor_tag;

	/* Copy data */
	pRetVal->data.resize(data.size());
	if (pRetVal->data.size())
	{
		memcpy(pRetVal->data.data(), data.data(), pRetVal->data.size());
	}
	return pRetVal;
}

QDVBStrm::QDVBStrm()
{
	stream_type = 0;
	elementary_PID = 0;
}

QDVBStrm::~QDVBStrm()
{
	for (int i=0; i<(int)descs.size(); i++)
	{
		QDVBDesc* pDesc = descs[i];
		delete pDesc;
	}
}

QDVBStrm* QDVBStrm::Duplicate()
{
	QDVBStrm* pRetVal = new QDVBStrm();

	pRetVal->elementary_PID = elementary_PID;
	pRetVal->stream_type = stream_type;
	for (int i=0; i<(int)descs.size(); i++)
	{
		pRetVal->descs.push_back(descs[i]->Duplicate());
	}

	return pRetVal;
}

